"""Model for Source Type 2 feeder input data"""

import copy
import itertools
import json
import logging
import os

import click

from jade.exceptions import InvalidParameter
from jade.utils.utils import ExtendedJSONEncoder
from PyDSS.common import ControllerType
from disco.cli.common import handle_existing_dir
from disco.enums import Placement, Scale, SimulationType
from disco.models.base import PyDSSControllerModel
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.models.time_series_analysis_model import TimeSeriesAnalysisModel
from disco.sources.base import (
    BaseSourceDataModel,
    BaseOpenDssModel,
    SOURCE_CONFIGURATION_FILENAME,
    DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS,
    DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS,
)
from .source_tree_2_model_inputs import SourceTree2ModelInputs


logger = logging.getLogger(__name__)


COMMON_OPTIONS = (
    click.option(
        "-f",
        "--feeders",
        default=["all"],
        multiple=True,
        show_default=True,
        help="feeders to add; use ('all',) to auto-detect and add all feeders",
    ),
    click.option(
        "-d",
        "--dc-ac-ratios",
        default=["all"],
        multiple=True,
        show_default=True,
        help="DC/AC ratios to add; use ('all',) to auto-detect and add all DC/AC ratios",
    ),
    click.option(
        "-s",
        "--scales",
        default=["all"],
        multiple=True,
        show_default=True,
        help="scales to add; use ('all',) to auto-detect and add all scales",
    ),
    click.option(
        "-p",
        "--placements",
        default=["all"],
        multiple=True,
        show_default=True,
        help="placements to add; use ('all',) to auto-detect and add all placements",
    ),
    click.option(
        "-d",
        "--deployments",
        default=["all"],
        multiple=True,
        show_default=True,
        help="deployments to add; use ('all',) to auto-detect and add all deployments",
    ),
    click.option(
        "-l",
        "--penetration-levels",
        default=["all"],
        multiple=True,
        show_default=True,
        help="penetration-levels to add; use ('all',) to auto-detect and add all penetration-levels",
    ),
    click.option(
        "-m",
        "--master-file",
        default="Master_noPV.dss",
        show_default=True,
        help="Master file for the OpenDSS model",
    ),
    click.option(
        "-F",
        "--force",
        help="overwrite existing directory",
        is_flag=True,
        default=False,
        show_default=True,
    ),
)


def common_options(func):
    for option in reversed(COMMON_OPTIONS):
        func = option(func)
    return func


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-o",
    "--output",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["output_dir"],
    show_default=True,
    help="output directory",
)
@click.pass_context
def snapshot(
    ctx,
    feeders,
    dc_ac_ratios,
    scales,
    placements,
    deployments,
    penetration_levels,
    master_file,
    force,
    start,
    output,
):
    """Transform input data for a snapshot simulation"""
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": start,
        "step_resolution": 900,
        "simulation_type": SimulationType.SNAPSHOT,
    }
    SourceTree2Model.transform(
        input_path=input_path,
        output_path=output,
        simulation_params=simulation_params,
        simulation_model=SnapshotImpactAnalysisModel,
        feeders=feeders,
        dc_ac_ratios=dc_ac_ratios,
        scales=scales,
        placements=placements,
        deployments=deployments,
        penetration_levels=penetration_levels,
        master_file=master_file,
    )
    print(f"Transformed data from {input_path} to {output} for Snapshot Analysis.")


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-e",
    "--end",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["end_time"],
    show_default=True,
    help="simulation end time",
)
@click.option(
    "-r",
    "--resolution",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["step_resolution"],
    type=int,
    show_default=True,
    help="simulation step resolution in seconds",
)
@click.option(
    "-o",
    "--output",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["output_dir"],
    show_default=True,
    help="output directory",
)
@click.option(
    "--pv-profile",
    default=None,
    type=str,
    help="profile to use for all PV Systems",
)
@click.pass_context
def time_series(
    ctx,
    feeders,
    dc_ac_ratios,
    scales,
    placements,
    deployments,
    penetration_levels,
    master_file,
    force,
    start,
    end,
    resolution,
    output,
    pv_profile,
):
    """Transform input data for a time series simulation"""
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": end,
        "step_resolution": resolution,
        "simulation_type": SimulationType.QSTS,
    }
    SourceTree2Model.transform(
        input_path=input_path,
        output_path=output,
        simulation_params=simulation_params,
        simulation_model=TimeSeriesAnalysisModel,
        feeders=feeders,
        dc_ac_ratios=dc_ac_ratios,
        scales=scales,
        placements=placements,
        deployments=deployments,
        penetration_levels=penetration_levels,
        master_file=master_file,
        pv_profile=pv_profile,
    )
    print(
        f"Transformed data from {input_path} to {output} for TimeSeries Analysis."
    )


class SourceTree2Model(BaseOpenDssModel):
    """Source Type 2 Feeder Model Inputs Class"""

    TRANSFORM_SUBCOMMANDS = {
        "snapshot": snapshot,
        "time-series": time_series,
    }

    def __init__(self, data):
        data = copy.deepcopy(data)
        self._path = data.pop("path")
        self._feeder = data.pop("feeder")
        self._master = data.pop("master", "Master_noPV.dss")
        self._dcac = data.pop("dcac")
        self._scale = data.pop("scale")
        self._placement = data.pop("placement")
        self._deployment = data.pop("deployment")
        self._penetration_level = data.pop("penetration_level")
        self._loadshape_directory = data.pop("loadshape_directory")
        self._opendss_directory = data.pop("opendss_directory")
        self._pv_locations = data.pop("pv_locations")
        self._metadata_directory = data.pop("metadata_directory", None)
        if data.pop("is_base_case"):
            self._name = self.make_base_case_name(self._feeder, self._dcac)
        else:
            self._name = self.make_name(
                self._feeder,
                self._dcac,
                self._scale,
                self._placement,
                self._deployment,
                self._penetration_level,
            )
        data.pop("deployment_file")
        assert not data, str(data)

    @property
    def dc_ac_ratio(self):
        return self._dcac

    @property
    def substation(self):
        return None

    @property
    def feeder(self):
        return self._feeder

    @property
    def loadshape_directory(self):
        return self._loadshape_directory

    @property
    def opendss_directory(self):
        return self._opendss_directory

    @property
    def master_file(self):
        return os.path.join(self._opendss_directory, self._master)

    @property
    def metadata_directory(self):
        return self._metadata_directory

    @property
    def name(self):
        return self._name

    @property
    def pv_locations(self):
        return self._pv_locations

    @property
    def pydss_controllers(self):
        return PyDSSControllerModel(
            controller_type=ControllerType.PV_CONTROLLER,
            name="volt-var",
        )

    @staticmethod
    def get_transform_subcommand(name):
        if name not in SourceTree2Model.TRANSFORM_SUBCOMMANDS:
            raise InvalidParameter(f"{name} is not supported")
        return SourceTree2Model.TRANSFORM_SUBCOMMANDS[name]

    @staticmethod
    def list_transform_subcommands():
        return sorted(list(SourceTree2Model.TRANSFORM_SUBCOMMANDS.keys()))

    @classmethod
    def transform(
        cls,
        input_path,
        output_path,
        simulation_model,
        simulation_params,
        feeders=("all",),
        dc_ac_ratios=("all",),
        scales=("all",),
        placements=("all",),
        deployments=("all",),
        penetration_levels=("all",),
        master_file="Master.dss",
        pv_profile=None,
    ):
        inputs = SourceTree2ModelInputs(input_path)

        if feeders == ("all",):
            feeders = inputs.list_feeders()
        if dc_ac_ratios == ("all",):
            dc_ac_ratios = inputs.list_dcac_ratios()
        elif not isinstance(dc_ac_ratios[0], float):
            dc_ac_ratios = [float(x) for x in dc_ac_ratios]
        if scales == ("all",):
            scales = inputs.list_scales()
        elif not isinstance(scales[0], Scale):
            scales = [Scale(x) for x in scales]
        if placements == ("all",):
            placements = inputs.list_placements()
        elif not isinstance(placements[0], float):
            placements = [Placement(x) for x in placements]

        config = []
        for feeder, dcac, scale, placement in itertools.product(
            feeders, dc_ac_ratios, scales, placements
        ):
            key = inputs.create_key(feeder, dcac, scale, placement)
            if deployments == ("all",):
                _deployments = inputs.list_deployments(key)
            else:
                _deployments = [int(x) for x in deployments]
            for deployment in _deployments:
                if penetration_levels == ("all",):
                    levels = inputs.list_penetration_levels(key, deployment)
                else:
                    levels = [int(x) for x in penetration_levels]
                for level in levels:
                    deployment_file = inputs.get_deployment_file(key, deployment, level)
                    data = {
                        "path": input_path,
                        "feeder": feeder,
                        "master": master_file,
                        "dcac": dcac,
                        "scale": scale.value,
                        "placement": placement.value,
                        "deployment": deployment,
                        "penetration_level": level,
                        "deployment_file": deployment_file,
                        "loadshape_directory": inputs.get_loadshape_directory(feeder),
                        "opendss_directory": inputs.get_opendss_directory(feeder),
                        "pv_locations": [deployment_file],
                        "is_base_case": False,
                    }
                    model = cls(data)
                    path = os.path.join(output_path, feeder)
                    out_deployment = model.create_deployment(
                        model.name, path, pv_profile=pv_profile
                    )
                    item = {
                        "deployment": out_deployment,
                        "simulation": simulation_params,
                        "name": model.name,
                        "model_type": simulation_model.__name__,
                    }
                    config.append(simulation_model.validate(item).dict())

        filename = os.path.join(output_path, SOURCE_CONFIGURATION_FILENAME)
        with open(filename, "w") as f_out:
            json.dump(config, f_out, indent=2, cls=ExtendedJSONEncoder)
        logger.info("Wrote config to %s", filename)

    @staticmethod
    def make_name(feeder, dcac, scale, placement, deployment, penetration_level):
        fields = (feeder, dcac, scale, placement, deployment, penetration_level)
        return "__".join([str(x) for x in fields])

    @staticmethod
    def make_base_case_name(feeder, dcac):
        fields = (feeder, dcac, -1, -1, -1)
        return "__".join([str(x) for x in fields])



"""Defines Snapshot Impact Analysis object."""

import os
import logging
import math
import re

import numpy as np
from pandas import DataFrame

from jade.common import JOBS_OUTPUT_DIR
from jade.jobs.results_aggregator import ResultsAggregator
from jade.utils.timing_utils import track_timing, TimerStatsCollector, Timer
from jade.utils.utils import dump_data

from PyDSS.pydss_results import PyDssResults, PyDssScenarioResults
from disco.analysis import Analysis, Input
from disco.exceptions import AnalysisRunException
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.utils.custom_type import CustomType


logger = logging.getLogger(__name__)


# TODO: remove once we have debugged the performance issues
TIMER_STATS = TimerStatsCollector()


class SnapshotImpactAnalysis(Analysis):
    """Snapshot impact analysis class with default values"""
    # Values are overridden by impact_analysis_inputs.toml in our workflow.
    INPUTS = [
        Input('over_voltage', CustomType(float), 1.05),
        Input('under_voltage', CustomType(float), 0.95),
        Input('over_voltage_conservative', CustomType(float), 1.05833),
        Input('under_voltage_conservative', CustomType(float), 0.91667),
        Input('line_overload_1', CustomType('percent'), 100),
        Input('line_overload_2', CustomType('percent'), 150),
        Input('transformer_overload_1', CustomType('percent'), 100),
        Input('transformer_overload_2', CustomType('percent'), 150),
    ]

    def __init__(self, feeder, *args, **kwargs):
        self._include_voltage_deviation = False
        super(SnapshotImpactAnalysis, self).__init__(*args, **kwargs)
        self._feeder = feeder
        TIMER_STATS.clear()

    def run(self, output, **kwargs):
        """Run snapshot impact analysis

        Parameters
        ----------
        output : directory containing job outputs

        """
        base_config = os.path.join(output, 'config.json')
        config = PyDssConfiguration.deserialize(base_config)
        results = ResultsAggregator.list_results(output)
        result_lookup = {x.name: x for x in results}
        jobs_output = os.path.join(output, JOBS_OUTPUT_DIR)

        for job in config.iter_feeder_jobs(self._feeder):
            if result_lookup[job.name].return_code != 0:
                logger.info("Skip failed job %s", job.name)
                continue
            self._run_job(config, job, jobs_output)

        TIMER_STATS.log_stats(clear=True)

    @track_timing(TIMER_STATS)
    def _run_job(self, config, job, output):
        simulation = config.create_from_result(job, output)
        results = PyDssResults(simulation.pydss_project_path)
        scenario = results.scenarios[0]

        # get base scenario, if there is one
        base_scenario = None
        self._include_voltage_deviation = job.model.include_voltage_deviation

        if self._include_voltage_deviation:
            base_case = job.model.base_case

            if base_case is None:
                # TODO: ideally, these should bubble upwards and be caught and logged
                # in one place (job-run)
                logger.error(
                    "Error: include-voltage-deviation flag enabled with no base case given: %s",
                    job.name
                )
                raise AnalysisRunException(
                    "Error: include-voltage-deviation flag enabled with no base case given: {}".format(
                        job.name,
                    )
                )

            if base_case != self._job_name:
                base_job = config.get_feeder_job(job.model.deployment.feeder, base_case)
                base_simulation = config.create_from_result(base_job, output)
                base_results = PyDssResults(base_simulation.pydss_project_path)
                base_scenario = base_results.scenarios[0]

        voltage_violations = self._run_voltage_violations(scenario, base_scenario)
        line_loading = self._run_line_loading(scenario)
        transformer_loading = self._run_transformer_loading(scenario)

        results = self._get_violations_for_job(job, voltage_violations, line_loading,
                                               transformer_loading)

        self._add_to_results('violations', results)

        # output to csv
        result_df = DataFrame(columns=results.keys())
        result_df.loc[0] = results

        filename = os.path.join(
            output,
            job.name,
            "snapshot-impact-analysis-job-post-process.csv",
        )

        result_df.to_csv(filename, index=False)

        try:
            data = scenario.read_feeder_head_info()
            filename = os.path.join(output, job.name, "FeederHeadInfo.json")
            dump_data(data, filename)
        except Exception:
            # This is expected if the version of PyDSS is wrong.
            logger.exception("read_feeder_head_info failed")
            raise

    @track_timing(TIMER_STATS)
    def _run_voltage_violations(self, scenario, base_scenario):
        """Run voltage violations from hosting capacity analysis

        Parameters
        ----------
        scenario : ValuesByPropertyAcrossElementsResults

        """
        # just one terminal, all phases
        bus_voltages = self._normalize_dataframe_values(scenario, 'Buses', 'puVmagAngle', 1,
                                                        mag_ang='mag')
        vmin, vmax, uv1, ov1, uv_count1, ov_count1, uv2, ov2, uv_count2, ov_count2 = \
            _check_voltage_violations(
                # TODO dynamically grab all phases, 1 terminal
                bus_voltages['Value'],
                self.get_input('over_voltage').current_value,
                self.get_input('under_voltage').current_value,
                self.get_input('over_voltage_conservative').current_value,
                self.get_input('under_voltage_conservative').current_value
            )

        total_pv_kw = None
        total_pv_pmpp = None
        total_load_kw = _get_total_load_kw(scenario.read_element_info_file('Loads'))
        pv_to_load_ratio = None

        if 'PVSystems' in scenario.list_element_classes():
            total_pv_kw = _get_total_pv_kw(scenario.read_element_info_file('PVSystems'))
            total_pv_pmpp = _get_total_pv_pmpp(scenario.get_full_dataframe('PVSystems', 'Pmpp'))
            pv_to_load_ratio = round(100 * total_pv_pmpp / max(0.00001, total_load_kw), 2)

        self.VIOLATION_FLAG_NAMES = {
            'uv1_flag': 'undervoltage_flag(limit={})'.format(self.get_input('under_voltage').current_value),
            'ov1_flag': 'overvoltage_flag(limit={})'.format(self.get_input('over_voltage').current_value),
            'uv2_flag': 'undervoltage_flag(limit={})'.format(self.get_input('under_voltage_conservative').current_value),
            'ov2_flag': 'overvoltage_flag(limit={})'.format(self.get_input('over_voltage_conservative').current_value),
            'lo1_flag': 'line_overloading_flag(limit={})'.format(self.get_input('line_overload_1').current_value),
            'lo2_flag': 'line_overloading_flag(limit={})'.format(self.get_input('line_overload_2').current_value),
            'to1_flag': 'xfmr_overloading_flag(limit={})'.format(self.get_input('transformer_overload_1').current_value),
            'to2_flag': 'xfmr_overloading_flag(limit={})'.format(self.get_input('transformer_overload_2').current_value),
        }

        self.VIOLATION_COUNT_NAMES = {
            'uv1_count': 'undervoltage_count(limit={})'.format(self.get_input('under_voltage').current_value),
            'uv2_count': 'undervoltage_count(limit={})'.format(self.get_input('under_voltage_conservative').current_value),
            'ov1_count': 'overvoltage_count(limit={})'.format(self.get_input('over_voltage').current_value),
            'ov2_count': 'overvoltage_count(limit={})'.format(self.get_input('over_voltage_conservative').current_value),
            'lo1_count': 'line_overloading_count(limit={})'.format(self.get_input('line_overload_1').current_value),
            'lo2_count': 'line_overloading_count(limit={})'.format(self.get_input('line_overload_2').current_value),
            'to1_count': 'xfmr_overloading_count(limit={})'.format(self.get_input('transformer_overload_1').current_value),
            'to2_count': 'xfmr_overloading_count(limit={})'.format(self.get_input('transformer_overload_2').current_value),
        }

        violations = {
            'pv_kw': total_pv_kw,
            'pv_pmpp': total_pv_pmpp,
            'peak_load': total_load_kw,
            'pv_to_load_ratio': pv_to_load_ratio,
            'min_voltage': vmin,
            'max_voltage': vmax,
            self.VIOLATION_FLAG_NAMES['uv1_flag']: uv1,
            self.VIOLATION_FLAG_NAMES['ov1_flag']: ov1,
            self.VIOLATION_FLAG_NAMES['uv2_flag']: uv2,
            self.VIOLATION_FLAG_NAMES['ov2_flag']: ov2,
            self.VIOLATION_COUNT_NAMES['uv1_count']: uv_count1,
            self.VIOLATION_COUNT_NAMES['ov1_count']: ov_count1,
            self.VIOLATION_COUNT_NAMES['uv2_count']: uv_count2,
            self.VIOLATION_COUNT_NAMES['ov2_count']: ov_count2,
        }

        # ensure that deviation indices are set (base scenarios)
        if self._include_voltage_deviation:
            violations['max_voltage_deviation'] = None
            violations['voltage_deviation_flag'] = None
            violations['voltage_deviation_count'] = None

        if base_scenario is not None:
            # TODO: If this will be used then we need to only get this once.
            base_voltages = self._normalize_dataframe_values(base_scenario, 'Buses', 'puVmagAngle',
                                                             1, mag_ang='mag')
            voltage_deviation, voltage_deviation_flag, voltage_deviation_count = _compare_voltages(
                base_voltages['Value'],
                bus_voltages['Value']
            )

            violations['max_voltage_deviation'] = voltage_deviation
            violations['voltage_deviation_flag'] = voltage_deviation_flag
            violations['voltage_deviation_count'] = voltage_deviation_count

        return violations

    @track_timing(TIMER_STATS)
    def _run_line_loading(self, scenario):
        """Run line loading from hosting capacity analysis

        Parameters
        ----------
        scenario : ValuesByPropertyAcrossElementsResults

        """
        # just one terminal, all phases

        lines_currents_dataframe = self._normalize_dataframe_values(scenario, "Lines", "Currents",
                                                                    1, convert=True)

        _, max_line_loading, lo1, lv_count1, lo2, lv_count2 = \
            _get_line_loading(
                lines_currents_dataframe,
                scenario.get_full_dataframe("Lines", "NormalAmps"),
                self._job_name,
                self.get_input('line_overload_1').current_value,
                self.get_input('line_overload_2').current_value
            )
        line_loading = {
            # 'line_loadings': line_loadings,
            'max_line_loading': max_line_loading,
            self.VIOLATION_FLAG_NAMES['lo1_flag']: lo1,
            self.VIOLATION_FLAG_NAMES['lo2_flag']: lo2,
            self.VIOLATION_COUNT_NAMES['lo1_count']: lv_count1,
            self.VIOLATION_COUNT_NAMES['lo2_count']: lv_count2,
        }

        return line_loading

    @track_timing(TIMER_STATS)
    def _run_transformer_loading(self, scenario):
        """Run transformer from hosting capacity analysis

        Parameters
        ----------
        scenario : ValuesByPropertyAcrossElementsResults

        """
        transformers_df = scenario.read_element_info_file("Transformers")
        transformers_phase_info_df = scenario.read_element_info_file("TransformersPhase")
        transformers_currents_df = self._normalize_dataframe_values(scenario, "Transformers",
                                                                    "Currents", 1, convert=True)
        transformers_normal_amps_df = scenario.get_full_dataframe("Transformers", "NormalAmps")

        _, max_xfmr_loading, to1, tv_count1, to2, tv_count2 = \
            _get_transformer_loading(
                transformers_df,
                transformers_phase_info_df,
                transformers_currents_df,
                transformers_normal_amps_df,
                self.get_input('transformer_overload_1').current_value,
                self.get_input('transformer_overload_2').current_value
            )

        transformer_loading = {
            # 'xfmr_loading_s': xfmr_loading_s,
            'max_xfmr_loading': max_xfmr_loading,
            self.VIOLATION_FLAG_NAMES['to1_flag']: to1,
            self.VIOLATION_FLAG_NAMES['to2_flag']: to2,
            self.VIOLATION_COUNT_NAMES['to1_count']: tv_count1,
            self.VIOLATION_COUNT_NAMES['to2_count']: tv_count2,
        }

        return transformer_loading

    @track_timing(TIMER_STATS)
    def _normalize_dataframe_values(self, scenario, class_name, property_name, terminal=None,
                                    convert=False, **kwargs):
        """Normalize dataframe columns to row values

        Parameters
        ----------
        scenario : ValuesByPropertyAcrossElementsResults
        class_name : str
        property_name : str
        terminal : int
        convert : bool
            (optional) convert values to Magnitude
        **kwargs : dict
            any extra named parameters (mag_ang, etc) to be passed to get_full_dataframe

        Returns
        -------
        results_dataframe : DataFrame

        """
        results_list = list()
        phase_terminal = None
        if terminal is not None:
            phase_terminal = re.compile(rf"[ABCN]{terminal}")

        df = scenario.get_full_dataframe(class_name, property_name, phase_terminal=phase_terminal,
                                         **kwargs)
        for column in df.columns:
            name = PyDssScenarioResults.get_name_from_column(column)
            value = df[column].values.item()
            result = {'Name': name, 'PhaseTerminal': column, 'Value': value}

            if convert:
                complex_number = complex(value)
                result['Value'] = math.sqrt(complex_number.real**2 + complex_number.imag**2)

            results_list.append(result)

        columns = ['Name', 'PhaseTerminal', 'Value']
        return DataFrame(results_list, columns=columns)

    @track_timing(TIMER_STATS)
    def _get_violations_for_job(self, job, voltage, line, transformer):
        """Returns results expected in output csv file

        Parameters
        ----------
        job : Job
        voltage : dict
            voltage violations
        line : dict
            line load
        transformer : dict
            transformer load

        Returns
        ------
        dict

        """
        result = {}

        project_data = job.model.deployment.project_data

        # Job Information
        result['feeder'] = job.model.deployment.feeder
        result['deployment'] = job.model.deployment.deployment_file
        result['placement'] = project_data.get("placement")
        result['sample'] = project_data.get("sample")
        result['penetration'] = project_data.get("penetration")

        # Voltage Violations
        result.update(voltage)

        # Line Loading
        result.update(line)

        # Transformer Loading
        result.update(transformer)

        self.VIOLATION_COMBINATIONS = {
            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_1').current_value,
                self.get_input('transformer_overload_1').current_value,
                self.get_input('over_voltage').current_value,
                self.get_input('under_voltage').current_value,
               ): ['to1_flag', 'lo1_flag', 'uv1_flag', 'ov1_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_2').current_value,
                self.get_input('transformer_overload_1').current_value,
                self.get_input('over_voltage').current_value,
                self.get_input('under_voltage').current_value,
            ): ['to2_flag', 'lo1_flag', 'uv1_flag', 'ov1_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_1').current_value,
                self.get_input('transformer_overload_2').current_value,
                self.get_input('over_voltage').current_value,
                self.get_input('under_voltage').current_value,
            ): ['to1_flag', 'lo2_flag', 'uv1_flag', 'ov1_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_2').current_value,
                self.get_input('transformer_overload_2').current_value,
                self.get_input('over_voltage').current_value,
                self.get_input('under_voltage').current_value,
            ): ['to2_flag', 'lo2_flag', 'uv1_flag', 'ov1_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_1').current_value,
                self.get_input('transformer_overload_1').current_value,
                self.get_input('over_voltage_conservative').current_value,
                self.get_input('under_voltage_conservative').current_value,
            ): ['to1_flag', 'lo1_flag', 'uv2_flag', 'ov2_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_2').current_value,
                self.get_input('transformer_overload_1').current_value,
                self.get_input('over_voltage_conservative').current_value,
                self.get_input('under_voltage_conservative').current_value,
            ): ['to2_flag', 'lo1_flag', 'uv2_flag', 'ov2_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_1').current_value,
                self.get_input('transformer_overload_2').current_value,
                self.get_input('over_voltage_conservative').current_value,
                self.get_input('under_voltage_conservative').current_value,
            ): ['to1_flag', 'lo2_flag', 'uv2_flag', 'ov2_flag'],

            '{}L_{}T_{}OV_{}UV_pass_flag'.format(
                self.get_input('line_overload_2').current_value,
                self.get_input('transformer_overload_2').current_value,
                self.get_input('over_voltage_conservative').current_value,
                self.get_input('under_voltage_conservative').current_value,
            ): ['to2_flag', 'lo2_flag', 'uv2_flag', 'ov2_flag'],
        }

        # iterate through the combination flags to determine status
        for key, value in self.VIOLATION_COMBINATIONS.items():
            result[key] = not _check_violation_flags(transformer, line, voltage,
                                                     self.VIOLATION_FLAG_NAMES[value[0]],
                                                     self.VIOLATION_FLAG_NAMES[value[1]],
                                                     self.VIOLATION_FLAG_NAMES[value[2]],
                                                     self.VIOLATION_FLAG_NAMES[value[3]],
                                                     )

        return result


# TODO: refactor to take in the scenario here instead of remaking it in analysis.py
@track_timing(TIMER_STATS)
def _check_voltage_violations(bus_voltages, ub1=1.05, lb1=0.95, ub2=1.05833, lb2=0.91667):
    vmax = max(bus_voltages)
    vmin = min(bus_voltages)
    uv_count1 = len([v for v in bus_voltages if v < lb1])
    ov_count1 = len([v for v in bus_voltages if v > ub1])
    uv_count2 = len([v for v in bus_voltages if v < lb2])
    ov_count2 = len([v for v in bus_voltages if v > ub2])

    ov1 = vmax > ub1
    uv1 = vmin < lb1
    ov2 = vmax > ub2
    uv2 = vmin < lb2

    return vmin, vmax, uv1, ov1, uv_count1, ov_count1, uv2, ov2, uv_count2, ov_count2


# TODO: refactor to take in the scenario here instead of remaking it in analysis.py
@track_timing(TIMER_STATS)
def _get_line_loading(line_currents_df, line_normalamps_df, deployment_name, ub1=1.0, ub2=1.5):
    line_loadings = {}
    lv_count1 = 0
    lv_count2 = 0

    # loop through unique line names
    for line in line_currents_df.Name.unique():
        current_mag = []
        # filter out only results with current line name
        line_currents = line_currents_df.query(f"Name == '{line}'")
        nphases = int(len(line_currents))
        # loop through line's magnitudes
        for _, magnitude in line_currents.iterrows():
            current_mag.append(magnitude['Value'] / line_normalamps_df.iloc[0][f"{line}__NormalAmps"])

        # TODO Kwami: how can this be empty?
        # deployment_name is only being passed in to debug this problem.
        if not current_mag:
            line_loadings[line] = 0.0
            logger.error("current_mag was empty, set line_loadings=0.0 "
                         "nphases=%s lencolumns=%s deployment_name=%s line=%s",
                         nphases, len(line_currents_df.keys()), deployment_name, line)
        else:
            line_loadings[line] = max(current_mag)
            if max(current_mag) > ub1:
                lv_count1 += 1
            if max(current_mag) > ub2:
                lv_count2 += 1

    max_line_loading = max(list(line_loadings.values()))
    lo1 = max_line_loading > ub1
    lo2 = max_line_loading > ub2

    return line_loadings, max_line_loading, lo1, lv_count1, lo2, lv_count2


# TODO: refactor to take in the scenario here instead of remaking it in analysis.py
@track_timing(TIMER_STATS)
def _get_transformer_loading(transformer_dataframe, highside_phase_conn_dataframe,
                            transformer_currents_dataframe, transformer_normalamps_dataframe,
                            ub1=1.0, ub2=1.5):
    xfmr_loading_s = {}
    tv_count1 = 0
    tv_count2 = 0

    # loop through unique line names
    for xfmr in transformer_currents_dataframe.Name.unique():
        current_mag = []
        transformer_name = xfmr.replace('Transformer.', '')
        # find transformer data in given dataframes
        transformer = transformer_dataframe.query(f"Name == '{transformer_name}'").iloc[0]
        transformer_phase = highside_phase_conn_dataframe.query(f"Transformer == '{xfmr}'").iloc[0]
        transformer_currents = transformer_currents_dataframe.query(f"Name == '{xfmr}'")

        # set variables necessary
        # TODO: really this shouldn't be necessary, since the dataframe will only have
        # the phases needed when passed in
        nwindings = transformer["NumWindings"]
        high_side_connection = transformer_phase["HighSideConnection"]
        num_phases = int(transformer_phase["NumPhases"])
        phase = len(transformer_currents)

        if num_phases == 3 and high_side_connection == "wye":
            phase = 4
        elif num_phases == 3 and high_side_connection == "delta":
            phase = 3

        for _, magnitude in transformer_currents.iterrows():
            normal_amps = transformer_normalamps_dataframe.iloc[0][f"{xfmr}__NormalAmps"]
            current_mag.append(magnitude['Value'] / normal_amps)

        xfmr_loading_s[xfmr] = max(current_mag)
        if max(current_mag) > ub1:
            tv_count1 += 1
        if max(current_mag) > ub2:
            tv_count2 += 1

    max_xfmr_loading = max(list(xfmr_loading_s.values()))
    to1 = max_xfmr_loading > ub1
    to2 = max_xfmr_loading > ub2

    return xfmr_loading_s, max_xfmr_loading, to1, tv_count1, to2, tv_count2


def _compare_voltages(base_voltage_df, voltage_df, limit=0.03):
    voltage_diff = base_voltage_df - voltage_df
    voltage_deviation_magnitude = abs(voltage_diff)
    voltage_deviation_count = len([f for f in voltage_deviation_magnitude if f > limit])
    voltage_deviation = np.max(voltage_deviation_magnitude)
    voltage_deviation_flag = voltage_deviation > limit
    return voltage_deviation, voltage_deviation_flag, voltage_deviation_count


def _get_total_pv_kw(pvsystems_df):
    return sum(pvsystems_df['kW'])


def _get_total_pv_pmpp(pmpp_df):
    return np.sum(pmpp_df.values)


def _get_total_load_kw(loads_df):
    return sum(loads_df['kW'])


def _check_violation_flags(transformer, line, voltage,
                           transformer_flag, line_flag, voltage_flag_1, voltage_flag_2):
    """Private helper function for getting pass flags

    Parameters
    ----------
    transformer : dict
        transformer load
    line : dict
        line load
    voltage : dict
        voltage violations
    transformer_flag : str
    line_flag : str
    voltage_flag_1 : str
    voltage_flag_2 : str

    Returns
    ------
    bool

    """
    return (transformer[transformer_flag] is not None and transformer[transformer_flag]) | \
           (line[line_flag] is not None and line[line_flag]) | \
           (voltage[voltage_flag_1] is not None and voltage[voltage_flag_1]) | \
           (voltage[voltage_flag_2] is not None and voltage[voltage_flag_2])



import logging

logger = logging.getLogger(__name__)


class FailingTestBisector:
    """Helper class to find the highest index of a sequence that passes a test."""

    def __init__(self, count):
        self._count = count
        self._highest_pass = -1
        self._lowest_fail = count
        self._num_iterations = 0

    def get_first_index(self):
        """Return the first index to test."""
        first_index = self._count // 2
        logger.debug("first_index=%s", first_index)
        return first_index

    def get_next_index(self, last_index, last_result_passed):
        """Return the next index to test and a bool to indicate whether that index is final.

        Parameters
        ----------
        last_index : int
        last_result_passed : bool

        Returns
        -------
        tuple
            next index to test, True if done

        Raises
        ------
        NoPassingIndex
            Raised if there is no passing index.

        """
        self._num_iterations += 1
        if last_result_passed:
            index, done = self._get_next_index_after_pass(last_index)
        else:
            index, done = self._get_next_index_after_failure(last_index)
        logger.debug(
            "last_index=%s passed=%s next_index=%s done=%s",
            last_index,
            last_result_passed,
            index,
            done,
        )

        if done:
            logger.info(
                "Highest passing index %s num_iterations=%s", last_index, self._num_iterations
            )

        return index, done

    def _get_next_index_after_pass(self, last_index):
        if last_index == self._lowest_fail:
            return last_index, True
        if last_index == self._count - 1:
            return last_index, True
        if last_index > self._highest_pass:
            self._highest_pass = last_index

        next_index = (last_index + self._lowest_fail) // 2
        if next_index == last_index:
            next_index += 1
        if next_index == self._lowest_fail:
            return last_index, True
        return next_index, False

    def _get_next_index_after_failure(self, last_index):
        if last_index == self._highest_pass:
            return last_index, True
        if last_index < self._lowest_fail:
            self._lowest_fail = last_index

        next_index = (self._highest_pass + last_index) // 2
        if next_index == -1:
            raise NoPassingIndex(f"no passing index count={self._count}")
        if next_index == self._highest_pass:
            return next_index, True
        return next_index, False


class NoPassingIndex(Exception):
    """Raised when there is no passing index."""



# default capacitor bank settings
DEFAULT_CAPACITOR_SETTINGS = {
    "capONdelay": 0,
    "capOFFdelay": 0,
    "capdeadtime": 0,
    "PTphase": "AVG",
    "cap_control": "voltage",
    "terminal": 1,
    "capON": None,  # Customize during voltage upgrades
    "capOFF": None  # Customize during voltage upgrades
}

# default Substation LTC settings
DEFAULT_SUBLTC_SETTINGS = {
    "winding": 2,
    "delay": 45,  # in seconds
    "band": 2,  # deadband in volts
    "properties_to_be_defined": [
        "winding",
        "ptratio",
        "band",
        "vreg",
        "delay",
    ],
    "vreg": None # Customize during voltage upgrades
}

# default regulator settings
DEFAULT_REGCONTROL_SETTINGS = {}
DEFAULT_REGCONTROL_SETTINGS["conn"] = "wye"
DEFAULT_REGCONTROL_SETTINGS["delay"] = 15  # in seconds
DEFAULT_REGCONTROL_SETTINGS["band"] = 2  # deadband in volts
DEFAULT_REGCONTROL_SETTINGS["properties_to_be_defined"] = [
    "winding",
    "ptratio",
    "band",
    "vreg",
    "delay",
]



from jade.utils.utils import load_data
from disco.pipelines.base import PipelineTemplate
from disco.pipelines.creator import (
    SnapshotPipelineCreator,
    TimeSeriesPipelineCreator,
    UpgradePipelineCreator
)
from disco.enums import SimulationType
from disco.sources.factory import make_source_model


PIPELINE_CREATOR_MAPPING = {
    SimulationType.SNAPSHOT: SnapshotPipelineCreator,
    SimulationType.TIME_SERIES: TimeSeriesPipelineCreator,
    SimulationType.UPGRADE: UpgradePipelineCreator
}


class PipelineCreatorFactory:
    """Factory class for getting pipeline creator"""
    
    @classmethod
    def create(cls, template_file):
        """Create pipeline creator"""
        template = PipelineTemplate(template_file)
        simulation_type = SimulationType(template.data["simulation_type"])
        pipeline_creator_class = PIPELINE_CREATOR_MAPPING[simulation_type]
        pipeline_creator = pipeline_creator_class(template_file)
        return pipeline_creator



import logging
import sys
from abc import ABC, abstractmethod

from filelock import SoftFileLock

from disco.storage.db import create_engine
from disco.storage.ingesters import OutputIngester, dump_storage_index
from disco.storage.outputs import get_simulation_output
from disco.storage.parsers import OutputParser

logger = logging.getLogger(__name__)


class PipelineStep(ABC):
    """Abstract pipeline step"""
    
    @staticmethod
    def execute(self):
        """Execute the pipeline step"""


class PipelineBase(ABC):
    """Abstract pipeline class"""
    
    def __init__(self, database):
        self.database = database

    @abstractmethod
    def run(self, data):
        """Entry point of running pipeline steps"""


class InitializationStep(PipelineStep):
    """A step for initializing and preparing data required by pipeline"""

    def execute(self, data):
        """Execute this step to initialize pipeline"""
        logger.info("Initializing data pipeline.")
        if not self.validate_task(data):
            task_name = data["task_name"]
            logger.error(f"Task '{task_name}' already exists, please try another --task-name.")
            sys.exit(1)
        return data

    def validate_task(self, data):
        """Return False if task_name already """
        if not data["database"]:
            return True
        
        engine = create_engine(data["database"])
        with engine.connect() as conn:
            queryset = conn.execute("SELECT name FROM task")
            existing_names = set([row[0] for row in queryset])
            if data["task_name"] in existing_names:
                return False
        return True


class OutputParsingStep(PipelineStep):
    """A step for parsing output of DISCO simulation/analysis """

    def __init__(self, output):
        self.output = get_simulation_output(output)
    
    def execute(self, data):
        """Parse desired data from output"""
        parser = OutputParser(
            task_name=data["task_name"],
            model_inputs=data["model_inputs"],
            notes=data["notes"],
        )
        result = parser.parse(output=self.output)
        return result


class ResultIngestionStep(PipelineStep):
    """A step for ingesting parsed result into database"""
    
    def __init__(self, database):
        self.database = database

    def execute(self, data):
        """Ingest parsed data into database"""
        logger.info("Ingesting results into database.")
        lock_file = self.database + ".lock"
        with SoftFileLock(lock_file=lock_file, timeout=3600):
            ingester = OutputIngester(database=self.database)
            indexes = ingester.ingest(data=data)
            return indexes


class FinalizatonStep(PipelineStep):
    """A step for closing pipeline"""
    
    def __init__(self, output):
        self.output = output
    
    def execute(self, data):
        """Store indexes of db storage into JSON file"""
        logger.info("Closing data ingestion pipeline.")
        dump_storage_index(output=self.output, indexes=data)


class StoragePipeline(PipelineBase):
    """Pipeline class for parsing and ingesting DISCO simulation/analysis results"""
    
    def run(self, data):
        """Execute pipeline steps in sequential order"""
        step1 = InitializationStep()
        valid_data = step1.execute(data=data)
        
        step2 = OutputParsingStep(output=data["output"])
        result = step2.execute(data=valid_data)
        
        step3 = ResultIngestionStep(database=data["database"])
        indexes = step3.execute(data=result)
        
        step4 = FinalizatonStep(output=data["output"])
        step4.execute(data=indexes)
        logger.info(f"Done! Tables were ingested into database - {data['database']}")



from email.policy import default
import logging
import sys

import click

from jade.models import (
    HpcConfig,
    LocalHpcConfig,
    SingularityParams,
)
from jade.utils.utils import dump_data, load_data

from disco.enums import SimulationType
from disco.pipelines.base import TemplateSection, TemplateParams
from disco.enums import AnalysisType
from disco.extensions.upgrade_simulation.upgrade_configuration import (
    DEFAULT_UPGRADE_COST_DB_FILE,
    DEFAULT_UPGRADE_PARAMS_FILE
)
from disco.pipelines.factory import PipelineCreatorFactory
from disco.pipelines.utils import get_default_pipeline_template, check_hpc_config
from disco.pydss.pydss_configuration_base import get_default_reports_file, get_default_exports_file
from disco.sources.factory import make_source_model


logger = logging.getLogger(__name__)

SIMULATION_TYPE_CHOICE = [
    SimulationType.SNAPSHOT.value,
    SimulationType.TIME_SERIES.value,
    SimulationType.UPGRADE.value
]


@click.group()
def create_pipeline():
    """Create JADE pipeline for DISCO analysis"""


@click.command()
@click.argument("inputs")
@click.option(
    "-T", "--task-name",
    type=click.STRING,
    required=True,
    help="The task name of the simulation/analysis",
)
@click.option(
    "-P", "--preconfigured",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
    help="Whether inputs models are preconfigured"
)
@click.option(
    "-s", "--simulation-type",
    type=click.Choice(SIMULATION_TYPE_CHOICE, case_sensitive=True),
    default=SimulationType.SNAPSHOT.value,
    show_default=True,
    help="Choose a DISCO simulation type"
)
@click.option(
    "--with-loadshape/--no-with-loadshape",
    type=click.BOOL,
    is_flag=True,
    default=None,
    help="Indicate if loadshape file used for Snapshot simulation."
)
@click.option(
    "--auto-select-time-points/--no-auto-select-time-points",
    is_flag=True,
    default=True,
    show_default=True,
    help="Automatically select the time point based on max PV-load ratio for snapshot "
         "simulations. Only applicable if --with-loadshape.",
)
@click.option(
    "-d", "--auto-select-time-points-search-duration-days",
    default=365,
    show_default=True,
    help="Search duration in days. Only applicable with --auto-select-time-points.",
)
@click.option(
    "-i", "--impact-analysis",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable impact analysis computations",
)
@click.option(
    "-h", "--hosting-capacity",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable hosting capacity computations",
)
@click.option(
    "-u", "--upgrade-analysis",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable upgrade cost computations"
)
@click.option(
    "-c", "--cost-benefit",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable cost benefit computations"
)
@click.option(
    "-p", "--prescreen",
    type=click.BOOL,
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable PV penetration level prescreening"
)
@click.option(
    "-t", "--template-file",
    type=click.STRING,
    required=False,
    default="pipeline-template.toml",
    show_default=True,
    help="Output pipeline template file"
)
@click.option(
    "-r",
    "--reports-filename",
    default=None,
    type=click.STRING,
    help="PyDSS report options. If None, use the default for the simulation type.",
)
@click.option(
    "-S",
    "--enable-singularity",
    is_flag=True,
    default=False,
    show_default=True,
    help="Add Singularity parameters and set the config to run in a container.",
)
@click.option(
    "-C",
    "--container",
    type=click.Path(exists=True),
    help="Path to container",
)
@click.option(
    "-D",
    "--database",
    type=click.Path(),
    default="results.sqlite",
    show_default=True,
    help="The path of new or existing SQLite database"
)
@click.option(
    "-l",
    "--local",
    is_flag=True,
    default=False,
    show_default=True,
    help="Run in local mode (non-HPC)."
)
def template(
    inputs,
    task_name,
    preconfigured,
    simulation_type,
    with_loadshape,
    auto_select_time_points,
    auto_select_time_points_search_duration_days,
    impact_analysis,
    hosting_capacity,
    upgrade_analysis,
    cost_benefit,
    prescreen,
    template_file,
    reports_filename,
    enable_singularity,
    container,
    database,
    local,
):
    """Create pipeline template file"""
    if hosting_capacity and impact_analysis:
        print("--impact-analysis and --hosting-capacity cannot both be enabled.")
        sys.exit(1)
    
    template = get_default_pipeline_template(simulation_type=simulation_type)
    template.data["task_name"] = task_name
    template.data["inputs"] = inputs
    template.data["database"] = database

    # model transformation
    if preconfigured:
        template.data["preconfigured"] = True
        template.remove_section(TemplateSection.MODEL)
    else:
        source_model = make_source_model(inputs)
        transform_defaults = source_model.get_transform_defaults()
        template.update_transform_params(transform_defaults)
    
    simulation_type = SimulationType(simulation_type)
    config_params = template.get_config_params(TemplateSection.SIMULATION)
    
    # snapshot special cases
    if simulation_type == SimulationType.SNAPSHOT:
        if prescreen:
            print("-p or --prescreen option has no effect on 'snapshot' pipeline, ignored!")
        
        if cost_benefit:
            print("--cost-benefit is not supported in snapshot simulations", file=sys.stderr)
            sys.exit(1)

        if with_loadshape is None:
            print("--with-loadshape option is required for Snapshot simulation.")
            sys.exit(1)

        config_params["with_loadshape"] = with_loadshape
        config_params["auto_select_time_points"] = auto_select_time_points
        config_params["auto_select_time_points_search_duration_days"] = \
            auto_select_time_points_search_duration_days
    
        template.update_config_params(config_params, TemplateSection.SIMULATION)

        if reports_filename is None:
            reports_filename = get_default_reports_file(simulation_type)
        template.update_reports_params(load_data(reports_filename))

    # time-series special cases
    if simulation_type == SimulationType.TIME_SERIES:
        if prescreen:
            template.remove_params(TemplateSection.SIMULATION, TemplateParams.CONFIG_PARAMS)
        else:
            template.remove_section(TemplateSection.PRESCREEN)
        
        if cost_benefit:
            if config_params["exports_filename"] is None:
                exports_filename = get_default_exports_file(
                    SimulationType.TIME_SERIES,
                    AnalysisType.COST_BENEFIT,
                )
                config_params["exports_filename"] = exports_filename
                template.update_config_params(config_params, TemplateSection.SIMULATION)
    
        if reports_filename is None:
            reports_filename = get_default_reports_file(simulation_type)
        template.update_reports_params(load_data(reports_filename))

    # upgrade special case
    if simulation_type == SimulationType.UPGRADE:
        config_params["cost_database"] = DEFAULT_UPGRADE_COST_DB_FILE
        config_params["params_file"] = DEFAULT_UPGRADE_PARAMS_FILE
        template.update_config_params(config_params, TemplateSection.SIMULATION)

    if impact_analysis:
        template.data["analysis_type"] = AnalysisType.IMPACT_ANALYSIS.value
    elif hosting_capacity:
        template.data["analysis_type"] = AnalysisType.HOSTING_CAPACITY.value
    elif upgrade_analysis:
        template.data["analysis_type"] = AnalysisType.UPGRADE_ANALYSIS.value
    elif cost_benefit:
        template.data["analysis_type"] = AnalysisType.COST_BENEFIT.value
    else:
        template.data["analysis_type"] = AnalysisType.NONE.value
        template.remove_section(TemplateSection.POSTPROCESS)
    
    if enable_singularity:
        singularity_params = SingularityParams(enabled=True, container=container)
        for section in template.data.values():
            if isinstance(section, dict) and "submitter-params" in section:
                section["submitter-params"]["singularity_params"] = singularity_params.dict()

    if local:
        for section in template.data.values():
            if isinstance(section, dict) and "submitter-params" in section:
                hpc_config = HpcConfig(hpc_type="local", hpc=LocalHpcConfig())
                section["submitter-params"]["hpc_config"] = hpc_config.dict()
                type_val = hpc_config.hpc_type.value
                section["submitter-params"]["hpc_config"]["hpc_type"] = type_val

    dump_data(template.data, filename=template_file)
    print(f"Pipeline template file created - {template_file}")


@click.command()
@click.argument("template-file")
@click.option(
    "-c", "--config-file",
    type=click.STRING,
    default="pipeline.json",
    show_default=True,
    help="Pipeline config file"
)
def config(template_file, config_file):
    """Create pipeline config file"""
    check_hpc_config(template_file)
    pipeline_creator = PipelineCreatorFactory.create(template_file=template_file)
    pipeline_creator.create_pipeline(config_file)
    print(f"Pipeline config file created - {config_file}")


create_pipeline.add_command(template)
create_pipeline.add_command(config)



import logging
import os

from jade.utils.utils import load_data

from disco.pydss.common import ConfigType
from disco.extensions.upgrade_simulation.upgrade_inputs import UpgradeInputs
from disco.pydss.pydss_configuration_base import PyDssConfigurationBase

logger = logging.getLogger(__name__)


DEFAULT_UPGRADE_COST_DB_FILE = os.path.join(
    os.path.dirname(__file__),
    "upgrades",
    "Generic_DISCO_cost_database_v2.xlsx"
)

DEFAULT_UPGRADE_PARAMS_FILE = os.path.join(
    os.path.dirname(__file__),
    "upgrades",
    "upgrade_parameters.toml"
)


class UpgradeConfiguration(PyDssConfigurationBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._pydss_inputs = self.get_default_pydss_config()

        # Customize pydss config
        if "enable_pydss_solve" in kwargs:
            self.enable_pydss_solve(kwargs["enable_pydss_solve"])

    @classmethod
    def auto_config(cls, inputs, **kwargs):
        if isinstance(inputs, str):
            inputs = UpgradeInputs(inputs)
        
        config = cls(**kwargs)
        for job in inputs.iter_jobs():
            config.add_job(job)
        
        return config

    def _serialize(self, data):
        data["pydss_inputs"] = self.serialize_pydss_inputs(self._pydss_inputs)

    @staticmethod
    def get_default_pydss_config():
        """Return the default PyDSS configuration.

        Returns
        -------
        dict

        """
        # this method not in use, simply return an empty dict.
        return {}

    def enable_pydss_solve(self, value: bool):
        self._pydss_inputs[ConfigType.SIMULATION_CONFIG]["default"]["enable_pydss_solve"] = value
        if value is True:
            message = "Enable PyDSS solve."
        else:
            message = "Disable PyDSS solve for upgrade simulation."
        logger.info(message)

    def get_pydss_controller_model(self, name):
        controller_config = self._pydss_inputs[ConfigType.CONTROLLER_CONFIG]
        for controller in controller_config:
            if controller["name"] != name:
                continue
            data = load_data(filename= controller["filename"])
            if name not in data:
                continue
            return data[name]
        raise ValueError(f"No controller named '{name}' configured.")

    def create_from_result(self, job, output_dir):
        cls = self.job_execution_class(job.extension)
        return cls.create(job, output=output_dir)



import os
import pathlib

import click

from jade.utils.run_command import check_run_command
from disco.pipelines.update_config import set_hc_postprocess_blocked_by
from disco.pipelines.utils import ensure_jade_pipeline_output_dir


@click.command()
@click.argument("commands-file")
def auto_config(commands_file):
    path = pathlib.Path(commands_file)
    commands = path.read_text().split("\n")
    for cmd in commands:
        cmd = cmd.strip()
        if not cmd:
            continue
        cmd = ensure_jade_pipeline_output_dir(cmd)
        check_run_command(cmd)
    
    if "pipeline-postprocess-auto-config.txt" in commands_file:
        set_hc_postprocess_blocked_by(commands_file)


if __name__ == "__main__":
    auto_config()



import fileinput
import logging
import os
import re
import shutil
from abc import ABC, abstractmethod
from pathlib import Path

from disco.enums import SimulationType, SimulationHierarchy
from disco.exceptions import AnalysisConfigurationException
from disco.models.base import OpenDssDeploymentModel
from disco.utils.dss_utils import comment_out_leading_strings

FORMAT_FILENAME = "format.toml"
TYPE_KEY = "type"

SOURCE_CONFIGURATION_FILENAME = "configurations.json"

DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS = {
    "output_dir": "snapshot-models",
    "start_time": "2020-04-15T14:00:00",
    "end_time": "2020-04-15T14:00:00",
    "simulation_type": SimulationType.SNAPSHOT.value,
}

DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS = {
    "output_dir": "time-series-models",
    "start_time": "2020-01-01T00:00:00",
    "end_time": "2020-01-02T00:00:00",
    "simulation_type": SimulationType.QSTS.value,
    "step_resolution": 900,
}

DEFAULT_UPGRADE_COST_ANALYSIS_PARAMS = {
    "output_dir": "upgrade-models",
    "start_time": "2020-01-01T00:00:00",
    "end_time": "2020-01-08T00:00:00",
    "simulation_type": SimulationType.SNAPSHOT.value,
}

DEFAULT_PV_DEPLOYMENTS_DIRNAME = "hc_pv_deployments"

logger = logging.getLogger(__name__)


class BaseSourceDataModel(ABC):
    """Base class for source data models"""

    @staticmethod
    @abstractmethod
    def get_transform_subcommand(name):
        """Return the click transform-model command.

        Returns
        -------
        click.command

        """

    @staticmethod
    @abstractmethod
    def list_transform_subcommands():
        """List the available click transform-model subcommands.

        Returns
        -------
        list
            List of click command objects

        """

    @classmethod
    @abstractmethod
    def transform(cls, config, simulation_model, output_path):
        """Transform the input data to a DISCO data model.

        Parameters
        ----------
        config : dict
        simulation_model : BaseAnalysisModel
        output_path : str

        """


class BaseOpenDssModel(BaseSourceDataModel, ABC):
    """Base model for a single OpenDSS configuration"""

    # Example line:
    # New Loadshape.Residential1234 npts=123456 minterval=5 mult=[file=../BuildingData/Dataset_12_34/Residential/RES1234/LoadProfiles/12345.csv]
    REGEX_LOAD_SHAPE_DATA_FILE = re.compile(r"file=([\.\w/\\-]+)")

    @property
    @abstractmethod
    def substation(self):
        """The substation name"""

    @property
    @abstractmethod
    def feeder(self):
        """The feeder name"""

    @property
    @abstractmethod
    def dc_ac_ratio(self):
        """The DC/AC ratio"""

    @property
    def kva_to_kw_rating(self):
        """The kva_to_kw_rating"""
        return 1.0

    @property
    @abstractmethod
    def name(self):
        """The name of the deployment"""

    @property
    @abstractmethod
    def loadshape_directory(self):
        """The directory containing load shapes"""

    @property
    @abstractmethod
    def opendss_directory(self):
        """The directory of OpenDSS model"""

    @property
    @abstractmethod
    def master_file(self):
        """Master file of OpenDSS model"""

    @property
    @abstractmethod
    def metadata_directory(self):
        """The directory of project metadata"""

    @property
    def project_data(self):
        """PyDSS controllers"""
        return {}

    @property
    @abstractmethod
    def pydss_controllers(self):
        """PyDSS controllers"""

    @property
    @abstractmethod
    def pv_locations(self):
        """PV systems file of OpenDSS model."""

    def _create_common_files(self, workspace, copy_load_shape_data_files, hierarchy):
        """Create files common to all deployments.

        Parameters
        ----------
        workspace : OpenDssFeederWorkspace
        copy_load_shape_data_files : bool

        """
        self._copy_files(
            src_dir=self.opendss_directory,
            dst_dir=workspace.opendss_directory,
            copy_load_shape_data_files=copy_load_shape_data_files,
        )
        if hierarchy == SimulationHierarchy.FEEDER and not os.path.exists(self.master_file):
            raise AnalysisConfigurationException(f"{self.master_file} is not present")

        if os.path.exists(self.master_file):
            # This may overwrite a file copied above.
            shutil.copyfile(self.master_file, workspace.master_file)
            # If you modify this list, compare the similar list in scripts/copy_smart_ds_dataset.py
            # The two locations may have different goals, and so do not share the same list reference.
            strings_to_remove = (
                "solve",
                "batchedit fuse",
                "new energymeter",
                "new monitor",
                "export monitors",
                "plot",
            )
            comment_out_leading_strings(workspace.master_file, strings_to_remove)

        if self.loadshape_directory is not None:
            self._copy_files(
                src_dir=self.loadshape_directory,
                dst_dir=workspace.loadshape_directory,
                copy_load_shape_data_files=copy_load_shape_data_files,
            )
        
        if self.metadata_directory is not None and os.path.exists(self.metadata_directory):
            self._copy_files(
                src_dir=self.metadata_directory,
                dst_dir=workspace.metadata_directory
            )

    def create_base_case(self, name, outdir, copy_load_shape_data_files=False):
        """Create a base case with no added PV.

        Parameters
        ----------
        name : str
            The job name
        outdir : str
            The base directory of opendss feeder model.
        copy_load_shape_data_files : bool

        Returns
        -------
        OpenDssDeploymentModel

        """
        workspace = OpenDssFeederWorkspace(outdir)
        if not os.path.exists(workspace.master_file):
            self._create_common_files(
                workspace,
                copy_load_shape_data_files,
                SimulationHierarchy.FEEDER,
            )

        deployment_file = Path(workspace.pv_deployments_directory) / (name + ".dss")
        rel_path = self._get_master_file_relative_path(deployment_file, Path(workspace.master_file))
        with open(deployment_file, "w") as fw:
            fw.write(f"Redirect {rel_path}\n")
            fw.write("\nSolve\n")

        return OpenDssDeploymentModel.validate(
            dict(
                deployment_file=str(deployment_file),
                substation=self.substation,
                feeder=self.feeder,
                dc_ac_ratio=self.dc_ac_ratio,
                directory=outdir,
                kva_to_kw_rating=self.kva_to_kw_rating,
                project_data=self.project_data,
                pydss_controllers=self.pydss_controllers,
            )
        )

    def create_substation_base_case(self, name, outdir, copy_load_shape_data_files=False):
        """Create a base case with no added PV.

        Parameters
        ----------
        name : str
            The job name
        outdir : str
            The base directory of opendss substation model.
        copy_load_shape_data_files : bool

        Returns
        -------
        OpenDssDeploymentModel

        """
        workspace = OpenDssSubstationWorkspace(outdir)
        if not os.path.exists(workspace.master_file):
            self._create_common_files(
                workspace,
                copy_load_shape_data_files,
                SimulationHierarchy.SUBSTATION,
            )

        deployment_file = Path(workspace.pv_deployments_directory) / (name + ".dss")
        rel_path = self._get_master_file_relative_path(deployment_file, Path(workspace.master_file))
        with open(deployment_file, "w") as fw:
            fw.write(f"Redirect {rel_path}\n")
            fw.write("\nSolve\n")

        return OpenDssDeploymentModel.validate(
            dict(
                deployment_file=str(deployment_file),
                substation=self.substation,
                feeder="None",
                dc_ac_ratio=self.dc_ac_ratio,
                directory=outdir,
                kva_to_kw_rating=self.kva_to_kw_rating,
                project_data=self.project_data,
                pydss_controllers=None,
            )
        )

    def create_deployment(
        self,
        name,
        outdir,
        hierarchy,
        pv_profile=None,
        copy_load_shape_data_files=False,
        strip_load_shape_profiles=False
    ):
        """Create the deployment.

        Parameters
        ----------
        name : str
            The deployment name
        outdir : str
            The base directory of opendss feeder model.
        hierarchy : SimulationHierarchy
        pv_profile : str
            Optional load shape profile name to apply to all PVSystems
        copy_load_shape_data_files : bool
        strip_load_shape_profiles : bool
            Strip load shape profiles from DSS models.

        Returns
        -------
        OpenDssDeploymentModel

        """
        workspace = OpenDssFeederWorkspace(outdir)
        if not os.path.exists(workspace.master_file):
            self._create_common_files(workspace, copy_load_shape_data_files, hierarchy)
        
        if strip_load_shape_profiles:
            self._strip_load_profiles_from_loads_file(workspace)
            self._strip_shape_files_from_master_file(workspace)

        deployment_file = self._create_deployment_file(
            name=name,
            workspace=workspace,
            hierarchy=hierarchy,
            pv_profile=pv_profile,
            strip_load_shape_profiles=strip_load_shape_profiles
        )
        if hierarchy == SimulationHierarchy.FEEDER:
            directory = outdir
            feeder = self.feeder
        else:
            directory = os.path.dirname(outdir)
            feeder = "None"
        return OpenDssDeploymentModel.validate(
            dict(
                deployment_file=deployment_file,
                substation=self.substation,
                feeder=feeder,
                dc_ac_ratio=self.dc_ac_ratio,
                directory=directory,
                kva_to_kw_rating=self.kva_to_kw_rating,
                project_data=self.project_data,
                pydss_controllers=self.pydss_controllers,
            )
        )

    @staticmethod
    def _copy_files(src_dir, dst_dir, exclude=None, copy_load_shape_data_files=False):
        """Copy files from src to dst directory.

        Parameters
        ----------
        src_dir : str
            Source directory
        dst_dir : str
            Destination directory
        exclude : list | str, optional
            Excluded file names from copy, by default None
        copy_load_shape_data_files : bool

        """
        if not exclude:
            exclude = []

        if isinstance(exclude, str):
            exclude = [exclude]

        for name in os.listdir(src_dir):
            if name in exclude or os.path.isdir(os.path.join(src_dir, name)):
                continue
            src_file = os.path.join(src_dir, name)
            dst_file = os.path.join(dst_dir, name)
            shutil.copyfile(src_file, dst_file)
            if os.path.splitext(dst_file)[1] in (".dss", ".txt"):
                if copy_load_shape_data_files:
                    BaseOpenDssModel.copy_any_load_shape_data_files(Path(src_file), Path(dst_file))
                else:
                    BaseOpenDssModel.make_data_file_references_absolute(
                        os.path.abspath(src_dir), dst_file
                )

    def _create_deployment_file(self, name, workspace, hierarchy, pv_profile=None, strip_load_shape_profiles=False):
        """Create deployment dss file.

        Parameters
        ----------
        name : str
            Name of deployment.
        workspace : OpenDssFeederWorkspace
            Instance of OpenDssFeederWorkspace
        hierarchy : SimulationHierarchy
        pv_profile : str | dict
            Optional load shape profile name to apply to PVSystems.
            If str, apply the name to all PVSystems.
            If dict, keys are PVSystem names and values are profile names.
        strip_load_shape_profiles: bool
            Strip load shape profiles from PVsystems if True, default False.
        """
        deployment_file = Path(workspace.pv_deployments_directory) / (name + ".dss")
        rel_path = self._get_master_file_relative_path(deployment_file, Path(workspace.master_file))
        if not self.pv_locations:
            with open(deployment_file, "w") as fw:
                fw.write(f"Redirect {rel_path}\n\n")
                fw.write("\nSolve\n")
            return str(deployment_file)

        regex1 = re.compile("yearly=[\w\.\-_]+")
        regex2 = re.compile(r"new pvsystem\.([^\s]+)")
        with open(deployment_file, "w") as fw, fileinput.input(self.pv_locations) as fr:
            if hierarchy == SimulationHierarchy.FEEDER:
                fw.write(f"Redirect {rel_path}\n\n")
            for line in fr:
                
                # Exclude load profiles for upgrades to reduce simulation time.
                if strip_load_shape_profiles:
                    match = regex1.search(line)
                    if match:
                        value = match.group(0)
                        line = " ".join(line.split(value)).strip() + "\n"
                
                elif pv_profile is not None:
                    lowered = line.lower()
                    if "new pvsystem" in lowered and "yearly" not in lowered:
                        if isinstance(pv_profile, str):
                            profile = pv_profile
                        else:
                            match = regex2.search(lowered)
                            assert match, lowered
                            pv_system = match.group(1)
                            if pv_system not in pv_profile:
                                raise Exception(f"no profile found for {pv_system}")
                            profile = pv_profile[pv_system]
                        if profile is not None:
                            line = line.strip() + f" yearly={profile}\n"
                fw.write(line)

            if hierarchy == SimulationHierarchy.FEEDER:
                fw.write("\nSolve\n")

        return str(deployment_file)

    @staticmethod
    def copy_any_load_shape_data_files(src_file, dst_file):
        """Copy any referenced load shape data file into the destination directory.

        Parameters
        ----------
        src_file : Path
        dst_file : Path

        """
        # Feeder-level has an "OpenDSS" directory but substation-level
        # doesn't, hence this hack.
        is_feeder_level = "OpenDSS" in str(dst_file)
        if is_feeder_level:
            profiles_dir = dst_file.parent.parent / "profiles"
        else:
            profiles_dir = dst_file.parent / "profiles"
        # TODO: This could be optimized in the future. We could combine the caller's
        # copy of src_file to dst_file with this rewrite.
        with fileinput.input(files=[dst_file], inplace=True) as f:
            for line in f:
                matches = BaseOpenDssModel.REGEX_LOAD_SHAPE_DATA_FILE.findall(line)
                replacements = {}
                for match in matches:
                    src_data_file_rel_path = Path(match)
                    src_data_file_path = (src_file.parent / src_data_file_rel_path).resolve()
                    name = src_data_file_path.name
                    dst_data_file_path = profiles_dir / name
                    if not dst_data_file_path.exists():
                        shutil.copyfile(src_data_file_path, dst_data_file_path)
                        logger.debug("Copied load shape data file %s", dst_data_file_path)
                    if is_feeder_level:
                        replacements[match] = str(Path("..") / "profiles" / name)
                    else:
                        replacements[match] = str(Path("profiles") / name)
                for old, new in replacements.items():
                    line = line.replace(old, new)
                print(line, end="")

    def _get_master_file_relative_path(self, deployment_file, master_path):
        return Path("..") / (master_path.relative_to(deployment_file.parent.parent))

    @staticmethod
    def make_data_file_references_absolute(src_dir, filename):
        """Change the path to any data file referenced in a .dss file to its
        absolute path."""

        def replace_func(match):
            path = os.path.normpath(match.group(1).replace("\\", "/"))
            new_path = "file=" + os.path.normpath(os.path.join(src_dir, path))
            return new_path

        with fileinput.input(files=[filename], inplace=True) as f_in:
            for line in f_in:
                line = re.sub(BaseOpenDssModel.REGEX_LOAD_SHAPE_DATA_FILE, replace_func, line)
                print(line, end="")
    
    @staticmethod
    def _strip_shape_files_from_master_file(workspace):
        """Exclude load and pv shapes from master DSS model file"""
        new_lines = []
        with open(workspace.master_file, "r") as f:
            for line in f.readlines():
                lowered = line.strip().lower()
                if lowered.startswith("!"):
                    new_lines.append(line)
                    continue
                elif "redirect loadshapes.dss" in lowered or "redirect pvshapes" in lowered:
                    new_lines.append("!" + line)
                else:
                    new_lines.append(line)
        
        with open(workspace.master_file, "w") as f:
            f.writelines(new_lines)

    @staticmethod
    def _strip_load_profiles_from_loads_file(workspace):
        """Exclude load profiles from loads DSS model file"""
        loads = os.path.basename(workspace.loads_file)
        loads_in_master = False
        with open(workspace.master_file, "r") as f:
            for line in f.readlines():
                if line.strip().startswith("!"):
                    continue
                if loads in line:
                    loads_in_master = True
                    break
        
        if not loads_in_master:
            return
        
        regex = re.compile("yearly=[\w\.\-_]+")
        new_lines = []
        with open(workspace.loads_file, "r") as f:
            for line in f.readlines():
                if not line.strip():
                    continue
                matched = regex.search(line)
                if not matched:
                    new_lines.append(line)
                    continue
                value = matched.group(0)
                new_line = " ".join(line.split(value)).strip()
                new_lines.append(new_line + "\n")

        with open(workspace.loads_file, "w") as f:
            f.writelines(new_lines)


class OpenDssSubstationWorkspace:
    """Defines a substation and all dependent OpenDSS files."""

    def __init__(self, substation_directory):
        self._substation_directory = substation_directory
        self._create_directories()

    def _create_directories(self):
        os.makedirs(self.substation_directory, exist_ok=True)
        os.makedirs(self.pv_deployments_directory, exist_ok=True)
        os.makedirs(self.profiles_directory, exist_ok=True)

    @property
    def opendss_directory(self):
        return self.substation_directory

    @property
    def profiles_directory(self):
        return os.path.join(self.substation_directory, "profiles")

    @property
    def pv_deployments_directory(self):
        return os.path.join(self.substation_directory, "PVDeployments")

    @property
    def master_file(self):
        return os.path.join(self.substation_directory, "Master.dss")

    @property
    def substation_directory(self):
        return self._substation_directory



class OpenDssFeederWorkspace:
    """Defines a feeder and all dependent OpenDSS files."""

    def __init__(self, feeder_directory):
        self._feeder_directory = feeder_directory
        self._create_directories()

    def _create_directories(self):
        os.makedirs(self.feeder_directory, exist_ok=True)
        os.makedirs(self.loadshape_directory, exist_ok=True)
        os.makedirs(self.opendss_directory, exist_ok=True)
        os.makedirs(self.pv_deployments_directory, exist_ok=True)
        os.makedirs(self.profiles_directory, exist_ok=True)
        os.makedirs(self.metadata_directory, exist_ok=True)

    @property
    def feeder_directory(self):
        return self._feeder_directory

    @property
    def loadshape_directory(self):
        return os.path.join(self.feeder_directory, "LoadShapes")

    @property
    def opendss_directory(self):
        return os.path.join(self.feeder_directory, "OpenDSS")

    @property
    def profiles_directory(self):
        return os.path.join(self.feeder_directory, "profiles")

    @property
    def pv_deployments_directory(self):
        return os.path.join(self.feeder_directory, "PVDeployments")

    @property
    def master_file(self):
        return os.path.join(self.opendss_directory, "Master.dss")

    @property
    def metadata_directory(self):
        return os.path.join(self.feeder_directory, "Metadata")

    @property
    def loads_file(self):
        return os.path.join(self.opendss_directory, "Loads.dss")



from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Set

from pydantic.v1 import BaseModel, Field, root_validator, validator

from jade.utils.utils import load_data

from disco.models.base import BaseAnalysisModel, PyDSSControllerModel


class PowerFlowGenericModel(BaseAnalysisModel):
    """Parameters for each job in a power-flow simulation"""

    name: str = Field(
        title="name",
        description="Unique name identifying the job",
    )
    opendss_model_file: str = Field(
        title="opendss_model_file",
        description="Path to file used load the simulation model files",
    )
    model_type: str = Field(
        title="model_type",
        description="Model type",
        default="PowerFlowGenericModel",
    )
    blocked_by: Set[str] = Field(
        title="blocked_by",
        description="Names of jobs that must finish before this job starts",
        default=set(),
    )
    estimated_run_minutes: Optional[int] = Field(
        title="estimated_run_minutes",
        description="Optionally advises the job execution manager on how long the job will run",
    )
    substation: Optional[str] = Field(
        title="substation",
        description="Substation for the job",
    )
    feeder: Optional[str] = Field(
        title="feeder",
        description="Feeder for the job",
    )
    pydss_controllers: List[PyDSSControllerModel] = Field(
        title="pydss_controllers",
        description="Apply these PyDSS controllers to each corresponding element type. If empty, "
        "use pf1.",
        default=[],
    )
    project_data: Dict = Field(
        title="project_data",
        description="Optional user-defined metadata for the job",
    )


class PowerFlowSimulationBaseModel(BaseModel):
    """Base model for power-flow simulations."""

    class Config:
        title = "PowerFlowSimulationBaseModel"
        anystr_strip_whitespace = True
        validate_assignment = True
        validate_all = True
        extra = "forbid"
        use_enum_values = False

    jobs: List[PowerFlowGenericModel] = Field(
        title="jobs",
        description="Jobs to run as part of the simulation.",
    )
    include_control_mode: bool = Field(
        title="include_control_mode",
        description="Include a control mode (such as volt-var controls) scenario for each job.",
        default=True,
    )
    include_pf1: bool = Field(
        title="include_pf1",
        description="Include a Power Factor 1 scenario for each job.",
        default=True,
    )

    @root_validator(pre=True)
    def check_job_names(cls, values):
        names = set()
        for job in values["jobs"]:
            if job["name"] in names:
                raise ValueError(f"{job['name']} is duplicated")
            names.add(job["name"])
        return values

    @root_validator(pre=False)
    def check_pydss_controllers(cls, values):
        if not values["jobs"]:
            raise ValueError("no jobs are defined")

        num_pydss_controllers = len(values["jobs"][0].pydss_controllers)
        if len(values["jobs"]) > 1:
            for job in values["jobs"][1:]:
                if len(job.pydss_controllers) != num_pydss_controllers:
                    raise ValueError("All jobs must have the same number of pydss_controllers.")
        return values

    @root_validator(pre=False)
    def check_scenarios(cls, values):
        if not values["include_pf1"] and not values["include_control_mode"]:
            raise ValueError(
                "At least one of 'include_pf1' and 'include_control_mode' must be set."
            )
        return values

    @classmethod
    def from_file(cls, filename: Path):
        """Return an instance of UpgradeCostAnalysisSimulationModel from a file

        Parameters
        ----------
        filename : Path

        """
        return cls(**load_data(filename))


class PowerFlowSnapshotSimulationModel(PowerFlowSimulationBaseModel):
    """Defines a snapshot power-flow simulation."""

    model_type: str = Field(default="PowerFlowSnapshotSimulationModel")
    start_time: datetime = Field(
        title="start_time",
        description="Start time of simulation. May be overridden by auto-time-point-selection.",
        default="2020-04-15 14:00:00",
    )

    @validator("model_type")
    def check_model_type(cls, val):
        if val != "PowerFlowSnapshotSimulationModel":
            raise ValueError(
                f"model_type must be 'PowerFlowSnapshotSimulationModel' instead of {val}"
            )
        return val


class PowerFlowTimeSeriesSimulationModel(PowerFlowSimulationBaseModel):
    """Defines a time-series power-flow simulation."""

    model_type: str = Field(default="PowerFlowTimeSeriesSimulationModel")
    start_time: datetime = Field(
        title="start_time",
        description="Start time of simulation.",
        default="2020-01-01 00:00:00",
    )
    end_time: datetime = Field(
        title="end_time",
        description="End time of simulation.",
        default="2020-12-31 23:45:00",
    )
    step_resolution: int = Field(
        title="step_resolution",
        description="Step resolution of simulation in seconds.",
        default=900,
    )

    @validator("model_type")
    def check_model_type(cls, val):
        if val != "PowerFlowTimeSeriesSimulationModel":
            raise ValueError(
                f"model_type must be 'PowerFlowTimeSeriesSimulationModel' instead of {val}"
            )
        return val



"""Implements functionality for QSTS inputs."""


from collections import namedtuple
import enum
import logging
import os
import re

from jade.jobs.job_inputs_interface import JobInputsInterface
from jade.exceptions import InvalidParameter
from jade.utils.utils import handle_file_not_found, handle_key_error
from disco.enums import get_placement_from_value, get_scale_from_value, \
    SCALE_MAPPING


logger = logging.getLogger(__name__)


class SourceTree2ModelInputs(JobInputsInterface):
    """Implements functionality for Source Tree 2 inputs."""

    _PV_DEPLOYMENTS = os.path.join("PVDeployments", "new")
    _REGEX_DCAC_RATIO = re.compile(r"DCAC([\d\.]+)")
    _KEYS_TYPE = namedtuple(
        "SourceTree2Keys", "feeder, dcac_ratio, scale, placement"
    )

    def __init__(self, base_directory):
        """Constructs SourceTree2ModelInputs.

        Parameters
        ----------
        base_directory : str

        """
        self._base = base_directory
        self._inputs_dir = os.path.join(self._base, "inputs")

        if not os.path.exists(self._base):
            raise InvalidParameter("inputs directory does not exist: {}"
                                   .format(self._base))


        self._parameters = self._parse_directories()

        logger.debug("Created %s at %s", self.__class__.__name__,
                     self._base)

    @property
    def base_directory(self):
        return self._base

    def create_key(self, feeder, dcac_ratio, scale, placement):
        """Create a parameter key for accessing deployments and penetration
        levels.

        Parameters
        ----------
        feeder : str
        dcac_ratio : float
        scale : Scale
        placement : Placement

        Returns
        -------
        namedtuple

        """
        return self._KEYS_TYPE._make((feeder, dcac_ratio, scale, placement))

    def get_available_parameters(self):
        """Return a dictionary containing all available parameters.

        Returns
        -------
        dict
            dictionary layout::

                {
                    (Feeder, DcacRatio, Scale, Placement) : {
                        deployments: Set(
                            penetration_levels,
                        )
                    }
                }

        """
        return self._parameters

    def get_deployment_file(self, key, deployment, penetration_level):
        """Get the path to an input deployment file for the given parameters.

        Parameters
        ----------
        key : namedtuple
        deployment : int
        penetration_level : int

        Returns
        -------
        str
            full path to file

        """
        filename = os.path.join(
            self._inputs_dir,
            key.feeder,
            self._PV_DEPLOYMENTS,
            self._get_dcac_directory(key.dcac_ratio),
            SCALE_MAPPING[key.scale],
            key.placement.value,
            str(deployment),
            f"PV_Gen_{deployment}_{penetration_level}.txt"
        )

        if not os.path.exists(filename):
            raise InvalidParameter(f"file does not exist: {filename}")
        return filename

    def get_dss_filename(self, feeder, filename):
        """Return the path to DSS file.

        Parameters
        ----------
        feeder : str
        filename : str

        Returns
        -------
        str
            path to file

        Raises
        ------
        InvalidParameter
            Raised if the file does not exist.

        """
        dss_filename = os.path.join(self._inputs_dir, feeder, "OpenDSS",
                                    filename)
        if not os.path.exists(dss_filename):
            raise InvalidParameter(f"file does not exist: {dss_filename}")

        return dss_filename

    def get_loadshape_directory(self, feeder):
        """Return the path to the load shape files.

        Parameters
        ----------
        feeder : str

        Returns
        -------
        str

        """
        return os.path.join(self._inputs_dir, feeder, "LoadShapes")

    def get_opendss_directory(self, feeder):
        """Return the path to the OpenDSS files.

        Parameters
        ----------
        feeder : str

        Returns
        -------
        str

        """
        return os.path.join(self._inputs_dir, feeder, "OpenDSS")

    def list_feeders(self):
        """List available feeders.

        Returns
        -------
        list of str
            available feeders

        """
        return self._list_available_params("feeder")

    def list_dcac_ratios(self):
        """List available DC/AC ratios.

        Returns
        -------
        list of float
            available DC/AC ratios

        """
        return self._list_available_params("dcac_ratio")

    def list_scales(self):
        """List available scales.

        Returns
        -------
        list of Scale
            available scales

        """
        return self._list_available_params("scale")

    def list_placements(self):
        """List available placements.

        Returns
        -------
        list of Placement
            available placements

        """
        return self._list_available_params("placement")

    @handle_key_error
    def list_deployments(self, key):
        """List available deployments.

        Parameters
        ----------
        key : namedtuple

        Returns
        -------
        list of int
            available deployments

        """
        deployments = list(self._parameters[key].keys())
        deployments.sort()
        return deployments

    @handle_key_error
    def list_penetration_levels(self, key, deployment):
        """List available penetration levels.

        Parameters
        ----------
        key : namedtuple
        deployment : int

        Returns
        -------
        list of int
            available penetration levels

        """
        levels = list(self._parameters[key][deployment])
        levels.sort()
        return levels

    def list_penetration_levels_by_step(self, key, deployment, step=10):
        """List available penetration levels by step.

        Parameters
        ----------
        key : namedtuple
        deployment : int
        step : int
            Must be a multiple of 5.

        Returns
        -------
        list of int
            available penetration levels

        Raises
        ------
        InvalidParameter
            Raised if key or deployment is not valid.

        """
        available_levels = self.list_penetration_levels(key, deployment)

        start = available_levels[0]
        stop = available_levels[-1] + 1

        # Convert to set for faster searches.
        available_levels = set(available_levels)

        levels = []
        for level in range(start, stop, step):
            if level not in available_levels:
                raise InvalidParameter(
                    "penetration level {} is not available for {}".format(
                        level, (key, deployment))
                )

            levels.append(level)

        return levels

    @staticmethod
    def _get_items_from_directory(path):
        """Basically just strips out files like README.md."""
        return [x for x in os.listdir(path)
                if os.path.isdir(os.path.join(path, x))]

    def _get_dcac_ratios_from_directory(self, path):
        ratios = []
        for item in self._get_items_from_directory(path):
            match = self._REGEX_DCAC_RATIO.search(item)
            if match:
                ratios.append(float(match.group(1)))
        return ratios

    @staticmethod
    def _get_dcac_directory(dcac_ratio):
        return "DCAC" + str(dcac_ratio)

    def _list_available_params(self, param):
        params = set()
        is_enum = False
        for key in self._parameters:
            value = key._asdict()[param]
            assert value is not None
            params.add(value)
            if isinstance(value, enum.Enum):
                is_enum = True

        key = None
        if is_enum:
            key = lambda x: x.value
        return sorted(list(params), key=key)

    @handle_file_not_found
    def _list_feeders(self):
        feeders = self._get_items_from_directory(self._inputs_dir)
        feeders.sort()
        return feeders

    @handle_file_not_found
    def _list_dcac_ratios(self, feeder):
        dcac_path = os.path.join(
            self._inputs_dir,
            feeder,
            self._PV_DEPLOYMENTS,
        )
        ratios = self._get_dcac_ratios_from_directory(dcac_path)

        ratios.sort()
        return ratios

    @handle_file_not_found
    def _list_scales(self, feeder, dcac_ratio):
        scale_path = os.path.join(
            self._inputs_dir,
            feeder,
            self._PV_DEPLOYMENTS,
            self._get_dcac_directory(dcac_ratio),
        )
        scales = [get_scale_from_value(x)
                  for x in self._get_items_from_directory(scale_path)]
        scales.sort(key=lambda x: x.value)
        return scales

    @handle_file_not_found
    def _list_placements(self, feeder, dcac, scale):
        placement_path = os.path.join(
            self._inputs_dir,
            feeder,
            self._PV_DEPLOYMENTS,
            self._get_dcac_directory(dcac),
            SCALE_MAPPING[scale],
        )
        placements = [get_placement_from_value(x)
                      for x in self._get_items_from_directory(
                          placement_path)]
        placements.sort(key=lambda x: x.value)
        return placements

    @handle_file_not_found
    def _list_deployments(self, feeder, dcac, scale, placement):
        deployment_path = os.path.join(
            self._inputs_dir,
            feeder,
            self._PV_DEPLOYMENTS,
            self._get_dcac_directory(dcac),
            SCALE_MAPPING[scale],
            placement.value,
        )
        deployments = self._get_items_from_directory(deployment_path)
        return sorted([int(x) for x in deployments])

    @handle_file_not_found
    def _list_penetration_levels(self, feeder, dcac, scale, placement,
                                 deployment):
        penetration_path = os.path.join(
            self._inputs_dir,
            feeder,
            self._PV_DEPLOYMENTS,
            self._get_dcac_directory(dcac),
            SCALE_MAPPING[scale],
            placement.value,
            str(deployment),
        )

        levels = []
        regex = re.compile(r"PV_Gen_{}_(\d+)\.txt".format(deployment))
        for filename in os.listdir(penetration_path):
            match = regex.search(filename)
            if match:
                levels.append(int(match.group(1)))

        return sorted(levels)

    def _parse_directories(self):
        data = {}
        for feeder in self._list_feeders():
            self._parse_dcacs(data, feeder)

        return data

    def _parse_dcacs(self, data, feeder):
        for dcac in self._list_dcac_ratios(feeder):
            self._parse_scales(data, feeder, dcac)

    def _parse_scales(self, data, feeder, dcac):
        for scale in self._list_scales(feeder, dcac):
            self._parse_placements(data, feeder, dcac, scale)

    def _parse_placements(self, data, feeder, dcac, scale):
        for placement in self._list_placements(feeder, dcac, scale):
            self._parse_deployments(data, feeder, dcac, scale, placement)

    def _parse_deployments(self, data, feeder, dcac, scale, placement):
        key = self._KEYS_TYPE(feeder, dcac, scale, placement)
        data[key] = {}
        for deployment in self._list_deployments(feeder, dcac, scale,
                                                 placement):
            data[key][deployment] = set()
            self._parse_penetrations(data[key][deployment], feeder, dcac,
                                     scale, placement, deployment)

    def _parse_penetrations(self, data, feeder, dcac, scale, placement,
                            deployment):
        for level in self._list_penetration_levels(feeder, dcac, scale,
                                                   placement, deployment):
            data.add(level)
